#| sawfish.wm.ext.pager -- Code for communicating with C pager
   $Id: pager.jl,v 1.3 2002/04/09 11:10:14 pfeiffer Exp $

   Copyright (C) 2002 Daniel Pfeiffer <occitan@esperanto.org>
   Copyright (C) 2000 Satyaki Das <satyaki@theforce.stanford.edu>
		      Ryan Lovett <ryan@ocf.Berkeley.EDU>
		      Andreas Buesching <crunchy@tzi.de>
		      Hakon Alstadheim

   This file is part of sawfish.

   sawfish is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   sawfish is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with sawfish; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

|#

(define-structure sawfish.wm.ext.pager

    (export pager

	    pager-change-depth
	    pager-goto
	    pager-move-window
	    pager-tooltip)

    (open rep
	  rep.io.processes
	  rep.system
	  sawfish.wm.colors
	  sawfish.wm.custom
	  sawfish.wm.ext.tooltips
	  sawfish.wm.fonts
	  sawfish.wm.misc
	  sawfish.wm.stacking
	  sawfish.wm.util.window-order
	  sawfish.wm.viewport
	  sawfish.wm.windows
	  sawfish.wm.workspace)

;;; Customization code contributed by Ryan Lovett <ryan@ocf.Berkeley.EDU>
  (defgroup pager "Pager")

  ;; closures for out of scope call
  (let ((state (lambda () (send-windows)))
	(resize (lambda () (send-size t)))
	(color (lambda () (send-colors))))

    (defcustom pager-show-ignored-windows nil
      "Show ignored windows"
      :type boolean
      :group pager
      :after-set state)

    (defcustom pager-show-all-workspaces t
      "Show all workspaces"
      :type boolean
      :group pager
      :after-set resize)

    (defcustom pager-workspaces-per-column 1
      "The number of workspaces per column"
      :type number
      :range (1)
      :depends pager-show-all-workspaces
      :group pager
      :after-set resize)

    (defcustom pager-stickies-on-all-workspaces t
      "Show (workspace) sticky windows on all their workspaces"
      :type boolean
      :depends pager-show-all-workspaces
      :group pager
      :after-set state)

    (defcustom pager-stickies-on-all-viewports t
      "Show (viewport) sticky windows on all their viewports"
      :type boolean
      :group pager
      :after-set state)

    (defcustom pager-shrink-factor 32
      "Each length in the pager is this many times shorter than the original"
      :type number
      :group pager
      :after-set resize)

    (defcustom pager-focus t
      "Whether Button1 focuses the clicked window"
      :type boolean
      :group pager)

    (defcustom pager-warp-cursor nil
      "Whether Button1 warps the cursor to the clicked point"
      :type boolean
      :group pager)

    (defcustom pager-color-background (get-color "#d8d8d8")
      "Pager background"
      :type color
      :group pager
      :after-set color)

    (defcustom pager-color-viewport (get-color "#f0f0f0")
      "Current viewport color"
      :type color
      :group pager
      :after-set color)

    (defcustom pager-color-window (get-color "#8080d0")
      "Windows"
      :type color
      :group pager
      :after-set color)

    (defcustom pager-color-focus-window (get-color "#58a8ff")
      "Window with input focus"
      :type color
      :group pager
      :after-set color)

    (defcustom pager-color-window-border (get-color "#505050")
      "Window borders"
      :type color
      :group pager
      :after-set color)

    (defcustom pager-color-viewport-divider (get-color "#e8e8e8")
      "Lines separating viewports"
      :type color
      :group pager
      :after-set color)

    (defcustom pager-color-workspace-divider (get-color "#202020")
      "Lines separating workspaces"
      :type color
      :group pager
      :after-set color))


  (defcustom pager-tooltips-enabled t
    "Enable display of window labels"
    :type boolean
    :group pager)

  (defvar pager-executable (concat sawfish-exec-directory "/pager"))

  ;; Remembers the number of workspaces...
  (define pager-num-workspaces)

  ;; Remembers the viewport dimensions...
  (define vp-rows)
  (define vp-columns)

  (define vp-width)
  (define vp-height)

  (define ws-width)
  (define ws-height)

  (define pager-width)
  (define pager-height)

  (define process nil)
  (define hooks
    '((after-move-hook . send-window)
      (after-resize-hook . send-window)
      (after-restacking-hook . send-windows)
      (enter-workspace-hook . send-viewport)
      (focus-in-hook . send-focus)
      (focus-out-hook . send-focus)
      (iconify-window-hook . send-windows)
      (map-notify-hook . send-windows)
      (shade-window-hook . send-window)
      (uniconify-window-hook . send-windows)
      (unmap-notify-hook . send-windows)
      (unshade-window-hook . send-window)
      (viewport-moved-hook . send-viewport)
      (viewport-resized-hook . send-size)
      (window-maximized-hook . send-window)
      (window-state-change-hook . send-size)
      (window-unmaximized-hook . send-window)
      (workspace-state-change-hook . send-size)))

;;; Internal utilities

  ;; This is just to scale the window co-ords and dimensions.
  (define-macro (scale val #!optional x up)
    (if x
	(if up
	    `(round (/ (* ,val (screen-width)) vp-width))
	  `(round (/ (* ,val vp-width) (screen-width))))
      (if up
	  `(round (/ (* ,val (screen-height)) vp-height))
	`(round (/ (* ,val vp-height) (screen-height))))))

  (define-macro (ws-offset ws #!optional x)
    (if x
	`(if pager-show-all-workspaces
	     (* vp-columns (screen-width)
		(quotient ,ws pager-workspaces-per-column))
	   0)
      `(if pager-show-all-workspaces
	   (* vp-rows (screen-height)
	      (mod ,ws pager-workspaces-per-column))
	 0)))

  ;; Given a location in the pager window, this translates into a location
  ;; in the actual workspaces. Returns a list of 3 numbers. The first is the
  ;; workspace number while the remaining two are the location relative to the
  ;; start of the workspace.
  (define (convert x y)
    (list (if pager-show-all-workspaces
	      (+ (quotient y ws-height)
		 (* pager-workspaces-per-column (quotient x ws-width)))
	    current-workspace)
	  (scale (mod x ws-width) 'x 'up)
	  (scale (mod y ws-height) () 'up)))

  (define (get-window-info w)
    (if (or (not (window-id w))
	    (not (window-mapped-p w))
	    (window-get w 'iconified)
	    (unless pager-show-ignored-windows (window-get w 'ignored)))
	'(0 0 0 0 0)
      (let* ((pos (window-position w))
	     (x (max (+ (car pos) viewport-x-offset) 0))
	     (y (max (+ (cdr pos) viewport-y-offset) 0))
	     (dim (window-frame-dimensions w))
;	     (shaded (if (window-get w 'shaded) 2))
	     (l1 (or (if pager-show-all-workspaces
			 (window-workspaces w))
		     (and pager-show-all-workspaces
			  pager-stickies-on-all-workspaces
			  (window-get w 'sticky)
			  (let loop ((i (cdr (workspace-limits)))
				     r)
			    (if (< i (car (workspace-limits)))
				r
			      (loop (1- i) `(,i ,@r)))))
		     (list current-workspace))))
	(if (and pager-stickies-on-all-viewports
		 (window-get w 'sticky-viewport)
		 (setq x (mod x (screen-width))
		       y (mod y (screen-height))))
	    (let loop ((i (1- (* vp-columns vp-rows)))
		       (l l1)
		       r)
	      (if (< i 0)
		  r
		(if l
		    (loop i (cdr l)
			  `((,(window-id w)
			     ,(scale (+ x (* (mod i vp-columns) (screen-width)) (ws-offset (car l) 'x)) 'x)
			     ,(scale (+ y (* (quotient i vp-columns) (screen-height)) (ws-offset (car l))))
			     ,(scale (min (car dim) (max (- (screen-width) x) 0)) 'x)
			     ,(scale (min (cdr dim) (max (- (screen-height) y) 0))))
			    ,@r))
		  (loop (1- i) l1 r))))
	  (mapcar (lambda (ws)
		    `(,(window-id w)
		      ,(scale (+ x (ws-offset ws 'x)) 'x)
		      ,(scale (+ y (ws-offset ws)))
		      ,(scale (min (car dim) (max (- (* (screen-width) vp-columns) x) 0)) 'x)
		      ,(scale (min (cdr dim) (max (- (* (screen-height) vp-rows) y) 0)))))
		  l1)))))

;;; Functions that talk to the C program

  (define-macro (send #!rest args)
    `(if process (format process ,@args)))
'(define-macro (send #!rest args)
    `(when process
	 (format standard-error "::")(format standard-error ,@args)(format process ,@args)))
  (define (send-colors)
    "Tells the C program to change the colors."
    (send "c%s\n"
	  (mapcar color-rgb
		  (list pager-color-background
			pager-color-viewport
			pager-color-window
			pager-color-focus-window
			pager-color-window-border
			pager-color-viewport-divider
			pager-color-workspace-divider))))

  ;; Sends window-id of focussed window (or 0 if none is focussed).
  (define (send-focus #!rest args)
    (declare (unused args))
    (send "f%d\n" (if (input-focus)
		      (window-id (input-focus))
		    0)))

  (define (send-size #!optional force init)
    "This tells the pager what the size of its window is going to be."
    (unless (and (not force)
		 (eql pager-num-workspaces
		      (let ((val (workspace-limits)))
			(1+ (- (cdr val) (car val)))))
		 (eql (car viewport-dimensions) vp-columns)
		 (eql (cdr viewport-dimensions) vp-rows))
      (setq pager-num-workspaces
	    (if pager-show-all-workspaces
		(let ((val (workspace-limits)))
		  (1+ (- (cdr val) (car val))))
	      1)
	    vp-columns (car viewport-dimensions)
	    vp-rows (cdr viewport-dimensions)
	    vp-width (quotient (screen-width) pager-shrink-factor)
	    vp-height (quotient (screen-height) pager-shrink-factor)
	    ws-width (* vp-columns vp-width)
	    ws-height (* vp-rows vp-height)
	    pager-width (* ws-width
			   (ceiling (/ pager-num-workspaces
				       pager-workspaces-per-column)))
	    pager-height (* ws-height
			    (min pager-workspaces-per-column
				 pager-num-workspaces)))
      (send "s%d %d %d %d %d %d\n"
	    vp-width vp-height
	    ws-width ws-height
	    pager-width pager-height)
      (or init (send-windows))))

  ;; Send the viewport that is in focus.
  (define (send-viewport #!rest args)
    (declare (unused args))
    (send "v%d %d\n"
	  (scale (+ viewport-x-offset (ws-offset current-workspace 'x)) 'x)
	  (scale (+ viewport-y-offset (ws-offset current-workspace))))
    ;; should send only stickies instead, and only depending on options
    (send-windows))

  ;; When only the size or shading of a window changes send only the data
  ;; pertaining to that window.
  (define (send-window w #!rest args)
    (declare (unused args))
    (if (or (if pager-show-all-workspaces
		(cdr (window-workspaces w)))
	    (if pager-stickies-on-all-viewports
		(window-get w 'sticky-viewport))
	    (and pager-show-all-workspaces
		 pager-stickies-on-all-workspaces
		 (window-get w 'sticky)))
	(send-windows)
      (send "w%s\n" (get-window-info w))))

  (define (send-windows #!rest args)
    "Tell the C program what to display.  For each window we send five
integers: window id, position and dimensions.

Putting extra spaces anywhere in between will have *bad* consequences
because my C routine is fragile."
    (declare (unused args))
    (send "W%s\n"
	  (mapcar get-window-info
		  (if pager-show-all-workspaces
		      (stacking-order)
		    (filter (lambda (w)
			      (let ((ws (window-workspaces w)))
				(or (null ws)
				    (member current-workspace ws))))
			    (stacking-order))))))



  (define (pager #!optional stop)
    "This function (re)starts the pager.
With a non-nil argument stops the pager instead."
    (when process
      (kill-process process)
      (setq process ()))
    (unless stop
      (setq process
	    (make-process () (lambda ()
			       (and process
				    (not (process-in-use-p process))
				    (setq process nil)))))
      (start-process process pager-executable)
      (send-colors)
      (send-size t t)
      (send-viewport)
      (send-focus)
      (condition-case err-info
	(mapc (lambda (hook)
		(add-hook (car hook) (symbol-value (cdr hook)) t))
	      hooks)
	(error
	 (format standard-error "pager: error adding hooks %s\n" err-info)))))

;;; Functions called from C program for 3 buttons and tooltips

  (define (pager-goto w x y)
    "Change to viewport and/or workspace where the user clicked on the pager."
    (let* ((loc1 (convert x y))
	   (loc2 (list (car loc1)
		       (quotient (cadr loc1) (screen-width))
		       (quotient (caddr loc1) (screen-height)))))
      (if (eql (car loc2) current-workspace)
	  (apply set-screen-viewport (cdr loc2))
	(apply select-workspace-and-viewport loc2))
      (if pager-warp-cursor
	  (warp-cursor (% (cadr loc1) (screen-width))
		       (% (caddr loc1) (screen-height))))
      (and pager-focus
	   (setq w (get-window-by-id w))
	   (set-input-focus w))))

  (define (pager-change-depth w)
    "Raise or lower the window clicked on in the pager."
    (if (setq w (get-window-by-id w))
	(raise-lower-window w)))

  (define (pager-move-window w x y)
    "Moves window with id ID to co-ordinates (X, Y) where (X, Y) is what the
pager thinks the position of the window is."
    ;; First do bounds checking...
    (setq x (if (< x 0) 0 (if (>= x (- pager-width 5)) (- pager-width 5) x))
	  y (if (< y 0) 0 (if (>= y (- pager-height 5)) (- pager-height 5) y))
	  w (get-window-by-id w))
    (let ((location (convert x y))
	  (cws (if w (window-workspaces w))))
      (move-window-to
       w
       (- (scale (mod x ws-width) 'x 'up)
	  viewport-x-offset)
       (- (scale (mod y ws-height) () 'up)
	  viewport-y-offset))
      ;; Don't change workspace if the window is sticky or present in
      ;; multiple workspaces or we are in single workspace mode
      (and w cws (null (cdr cws)) (not (eql (car location) (car cws)))
	   pager-show-all-workspaces
	   (let* ((was-focused (eq (input-focus) w))
		  (orig-space (if (window-in-workspace-p w current-workspace)
				  current-workspace
				(car (window-workspaces w))))
		  (new-space (workspace-id-from-logical (car location))))
	     (when orig-space
	       (copy-window-to-workspace w orig-space new-space nil)
	       (if (eql orig-space current-workspace)
		   (delete-window-instance w))
	       (send-window w)
	       (move-window-to-workspace w orig-space new-space was-focused))))
      (send-window w)))

  (define (pager-tooltip #!optional id)
    "Show a tooltip for window ID, or remove it if no ID given."
    (when pager-tooltips-enabled
      (if (and id
	       (setq id (get-window-by-id id))
	       (setq id (window-name id)))
	  (let ((te tooltips-enabled)
		(tooltips-enabled t))
	    (display-tooltip-after-delay
	     (if te
		 (concat id "\n\n"
			 (_ "Button1-Click  select viewport (and optionally window)
Button2-Click  raise/lower window
Button3-Move   drag window"))
	       id)))
	(remove-tooltip)))))
